@@ -46,7 +46,7 @@ class Defer(object):
     If a future is attached during execution, set_running_or_notify_cancelled will
     not be invoked, but set_result will.
     """
-    
+
     def __init__(self, callable_, *args, **kwargs):
         self.callable_ = callable_
         self.args = args
@@ -82,29 +82,34 @@ class Defer(object):
             else:
                 future.set(rv)
 
+def _renew_undefer(self, key, ttl, kw, value):
+    self = self()
+    self.client.renew(key, ttl, **(kw or {}))
+    return value.undefer()
+
 _global_cleanup_tasks = []
 
 class AsyncCacheWriterThreadPool(ThreadPool):
     class AsyncCacheWriterThread(ThreadPool.Process):
         pass
     Process = AsyncCacheWriterThread
-    
+
     def __init__(self, workers):
         if ThreadPool is multiprocessing.pool.ThreadPool:
-            # This patches ThreadPool, which is broken when instanced 
+            # This patches ThreadPool, which is broken when instanced
             # from inside a DummyThread (happens after forking)
             current = multiprocessing.dummy.current_process()
             if not hasattr(current, '_children'):
                 current._children = weakref.WeakKeyDictionary()
-        
+
         ThreadPool.__init__(self, workers)
 
 class AsyncCacheWriterPool:
-    def __init__(self, size, workers, client, overflow = False, cleanup_cycles = 500, 
+    def __init__(self, size, workers, client, overflow = False, cleanup_cycles = 500,
             defer_threadpool = None, writer_threadpool = None):
-        
+
         self.client = client
-        self.logger = logging.getLogger("chorde")
+        self.logger = logging.getLogger("chorde.async")
         self.size = size
         self.workers = workers
         self._spawnlock = threading.Lock()
@@ -120,7 +125,7 @@ class AsyncCacheWriterPool:
         else:
             self._writer_threadpool_factory = None
             self._writer_threadpool = writer_threadpool
-        
+
         # queueset holds the values to be written, associated
         # by key, providing some write-back coalescense in
         # high-load environments
@@ -200,13 +205,13 @@ class AsyncCacheWriterPool:
                 except:
                     self.logger.error("Error in background cache refresh", exc_info=True)
                     value = _NONE
-            
+
             if value is _NONE or value is NONE:
                 # undefer probably decided not to compute anything (or an error arose, whatever)
                 if deferred is not _NONE:
                     deferred.done()
                 return
-                
+
             elif value is _DELETE:
                 try:
                     self.client.delete(key)
@@ -218,25 +223,25 @@ class AsyncCacheWriterPool:
                     self.client.renew(key, ttl, **(kw or {}))
                 except:
                     self.logger.error("Error renewing key", exc_info=True)
-            
+
             elif value is _EXPIRE:
                 try:
                     self.client.expire(key)
                 except:
                     self.logger.error("Error expiring key", exc_info=True)
-            
+
             elif value is _CLEAR:
                 try:
                     self.client.clear()
                 except:
                     self.logger.error("Error clearing cache", exc_info = True)
-            
+
             elif value is _PURGE:
                 try:
                     self.client.purge()
                 except:
                     self.logger.error("Error purging cache", exc_info = True)
-            
+
             else:
                 try:
                     self.client.put(key, value, ttl, **(kw or {}))
@@ -253,13 +258,13 @@ class AsyncCacheWriterPool:
             else:
                 kev = None
             del w
-            
+
             if thread_id is not None and thread_id not in map(operator.itemgetter(0), self.workset.values()):
                 try:
                     self.threadset.remove(thread_id)
                 except KeyError:
                     pass
-            
+
             ev.set()
             if kev is not None:
                 kev.set()
@@ -277,7 +282,7 @@ class AsyncCacheWriterPool:
                     task()
         except:
             self.logger.error("Error during cleanup", exc_info = True)
-        
+
     @property
     def capacity(self):
         return self.size
@@ -302,7 +307,7 @@ class AsyncCacheWriterPool:
         return rv
 
     def enqueue(self, key, value, ttl=None, **kw):
-        if (thread.get_ident() in self.threadset 
+        if (thread.get_ident() in self.threadset
                  or (hasattr(self.defer_threadpool, 'in_worker') and self.defer_threadpool.in_worker()) ):
             # Oops, recursive call, bad idea
             # Run inline
@@ -372,6 +377,17 @@ class AsyncCacheWriterPool:
                     else:
                         # Why not
                         delayed = functools.partial(future.set, queue_value)
+            else:
+                queue_value = queueset.get(key)
+                if queue_value is not None:
+                    queue_value, queue_ttl, queue_kwargs = queue_value
+                    if queue_value is _RENEW:
+                        # Turn renew + put serial operations into a single renew-then-put defer
+                        queueset[key] = (
+                            Defer(_renew_undefer, self._wself, key, queue_ttl, queue_kwargs, value),
+                            ttl,
+                            (kw or None)
+                        )
         else:
             # New one wins, we just have to chain any queued future
             queue_value = queueset.get(key)
@@ -406,11 +422,11 @@ class AsyncCacheWriterPool:
                 # defers go in concurrently, only the first will be invoked,
                 # instead of the last - the first cannot be canceled after all,
                 # and we want to only invoke one. So no choice.
-                
+
                 # ...if the defer has a future attached
                 future = getattr(value, 'future', None)
                 if future is not None and hasattr(future, 'add_done_callback'):
-                    # we do have to hook into the future though... 
+                    # we do have to hook into the future though...
                     working = workset.get(key)
                     if working is not None:
                         working = working[1]
@@ -431,36 +447,67 @@ class AsyncCacheWriterPool:
         else:
             delayed = self._coalesce(key, value, ttl, **kw)
         return delayed
-    
+
     def waitkey(self, key, timeout=None):
-        if thread.get_ident() in self.threadset:
+        tid = thread.get_ident()
+        if (tid in self.threadset
+                or (hasattr(self.defer_threadpool, 'in_worker') and self.defer_threadpool.in_worker())):
             # Oops, recursive call, bad idea
             return
-        elif timeout is None:
-            while self.contains(key):
-                ev = self.workset.get(key)
+
+        tidt = (tid,)
+        queueset = self.queueset
+        workset_get = self.workset.get
+        done_event = self.done_event
+        sleep = time.sleep
+        busyloop = 0
+
+        if timeout is None:
+            # contains inlined for speed in code below
+            while (key in queueset or workset_get(key, tidt)[0] != tid):
+                ev = workset_get(key)
                 if ev is not None:
                     ev = ev[2]
-                elif self.contains(key):
-                    ev = self.done_event
+                elif (key in queueset or workset_get(key, tidt)[0] != tid):
+                    ev = done_event
                 if ev is not None:
-                    ev.wait(1.0)
+                    done = ev.wait(1.0)
+                    if not done:
+                        busyloop = 0
+                    else:
+                        busyloop += 1
+                        if busyloop > 5:
+                            if ev is done_event and busyloop > 20 and ev.isSet():
+                                ev.clear()
+                            else:
+                                sleep(0.05)
                 else:
                     break
         else:
-            tfin = time.time() + timeout
-            while self.contains(key) and tfin >= time.time():
-                ev = self.workset.get(key)
+            # contains inlined for speed in code below
+            time_time = time.time
+            tfin = time_time() + timeout
+            while (key in queueset or workset_get(key, tidt)[0] != tid) and tfin >= time_time():
+                ev = workset_get(key)
                 if ev is not None:
                     ev = ev[2]
-                elif self.contains(key):
-                    ev = self.done_event
+                elif (key in queueset or workset_get(key, tidt)[0] != tid):
+                    ev = done_event
                 if ev is not None:
-                    ev.wait(min(1.0, timeout))
+                    done = ev.wait(min(1.0, timeout))
+                    if not done:
+                        busyloop = 0
+                    else:
+                        busyloop += 1
+                        if busyloop > 5:
+                            if ev is done_event and busyloop > 20 and ev.isSet():
+                                ev.clear()
+                            else:
+                                sleep(0.05)
                 else:
                     break
-                timeout = tfin - time.time()
-    
+                timeout = tfin - time_time()
+
     def getTtl(self, key, default = None):
         # Speeding up things - we have the same signature and semantics
         self.getTtl = getTtl = self.queueset.get
@@ -471,7 +518,7 @@ class AsyncCacheWriterPool:
         if key in self.queueset:
             return True
         else:
-            # Exclude keys being computed by this thread, such calls 
+            # Exclude keys being computed by this thread, such calls
             # want the async wrapper out of their way
             tid = thread.get_ident()
             return self.workset.get(key, (tid,))[0] != tid
@@ -538,37 +585,37 @@ class AsyncWriteCacheClient(BaseCacheClient):
         self.spawning_lock = threading.Lock()
         self._defer_threadpool = defer_threadpool or threadpool
         self._writer_threadpool = writer_threadpool or threadpool
-        
+
     def assert_started(self):
         if self.writer is None:
             with self.spawning_lock:
                 if self.writer is None:
                     self.writer = AsyncCacheWriterPool(
-                        self.writer_queue_size, 
+                        self.writer_queue_size,
                         self.writer_workers,
                         self.client,
                         self.overflow,
                         defer_threadpool = self._defer_threadpool,
                         writer_threadpool = self._writer_threadpool)
-    
+
     def is_started(self):
         return self.writer is not None
-    
+
     def start(self):
         if self.writer is not None:
             raise AssertionError, "Starting AsyncCacheClient twice"
         self.assert_started()
-    
+
     def stop(self, abort_tasks=False):
         if self.writer is not None:
             if not abort_tasks:
                 self.writer.join()
             self.writer.terminate()
-    
+
     @property
     def async(self):
         return True
-    
+
     @property
     def capacity(self):
         return (self.client.capacity, self.writer.capacity if self.writer is not None else 0)
@@ -580,11 +627,11 @@ class AsyncWriteCacheClient(BaseCacheClient):
     def put(self, key, value, ttl, **kw):
         self.assert_started()
         self.writer.put(key, value, ttl, **kw)
-    
+
     def renew(self, key, ttl, **kw):
         self.assert_started()
         self.writer.renew(key, ttl, **kw)
-    
+
     def delete(self, key):
         self.assert_started()
         self.writer.delete(key)
@@ -603,9 +650,9 @@ class AsyncWriteCacheClient(BaseCacheClient):
         """
         if self.is_started():
             self.writer.purge()
-    
-    def getTtl(self, key, default = NONE, 
-            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR, 
+
+    def getTtl(self, key, default = NONE,
+            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
             NONE = NONE, _NONE = _NONE,
             hasattr = hasattr,
             **kw):
@@ -635,12 +682,12 @@ class AsyncWriteCacheClient(BaseCacheClient):
 
             # Check pending clear - after checking the queue for sorted semantics
             if writer._contains(_CLEAR):
-                # Well, 
+                # Well,
                 if default is NONE:
                     raise CacheMissError, key
                 else:
                     return default, -1
-        
+
         # Ok, read the cache then
         value, ttl = self.client.getTtl(key, default, **kw)
         if ettl is not None:
@@ -651,7 +698,7 @@ class AsyncWriteCacheClient(BaseCacheClient):
             return value, ttl
 
     def get(self, key, default = NONE, ttl_skip = None,
-            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR, 
+            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
             NONE = NONE, _NONE = _NONE,
             hasattr = hasattr,
             **kw):
@@ -678,17 +725,17 @@ class AsyncWriteCacheClient(BaseCacheClient):
 
             # Check pending clear - after checking the queue for sorted semantics
             if writer._contains(_CLEAR):
-                # Well, 
+                # Well,
                 if default is NONE:
                     raise CacheMissError, key
                 else:
                     return default
-        
+
         # Ok, read the cache then
         return self.client.get(key, default, ttl_skip = ttl_skip, **kw)
 
-    def getTtlMulti(self, keys, default = NONE, 
-            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR, 
+    def getTtlMulti(self, keys, default = NONE,
+            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
             NONE = NONE, _NONE = _NONE,
             hasattr = hasattr,
             **kw):
@@ -728,7 +775,7 @@ class AsyncWriteCacheClient(BaseCacheClient):
                 for key in keys:
                     yield key, default_rv
                 return
-        
+
         # Ok, read the cache then
         ettlget = ettl.get
         for key, (value, ttl) in self.client.getTtlMulti(keys, default, **kw):
@@ -736,7 +783,7 @@ class AsyncWriteCacheClient(BaseCacheClient):
             yield key, (value, ttl)
 
     def getMulti(self, keys, default = NONE, ttl_skip = None,
-            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR, 
+            _DELETE = _DELETE, _EXPIRE = _EXPIRE, _RENEW = _RENEW, _CLEAR = _CLEAR,
             NONE = NONE, _NONE = _NONE,
             hasattr = hasattr,
             **kw):
@@ -792,11 +839,11 @@ class AsyncWriteCacheClient(BaseCacheClient):
             return
         else:
             return self.client.promote(key, *p, **kw)
-    
+
     def wait(self, key, timeout = None):
         if self.is_started() and self.writer.contains(key):
             self.writer.waitkey(key, timeout)
-    
+
     def contains(self, key, ttl = None, **kw):
         if self.is_started():
             if self.writer.contains(key):
@@ -829,10 +876,10 @@ except ImportError:
         "using pure-python version which is not atomic and requires"
         "explicit synchronization. Decreased performance will be noticeable")
     del warnings
-    
+
     class ExceptionWrapper(object):  # lint:ok
         __slots__ = ('value',)
-    
+
         def __init__(self, value):
             self.value = value
 
@@ -843,15 +890,15 @@ except ImportError:
 
     class Future(object):  # lint:ok
         __slots__ = (
-            '_cb', '_value', '_logger', '_running', '_cancel_pending', '_cancelled', '_done_event', 
+            '_cb', '_value', '_logger', '_running', '_cancel_pending', '_cancelled', '_done_event',
             '_lock', '__weakref__',
         )
-        
+
         def __init__(self, logger = None):
             self._cb = []
             self._logger = logger
             self._lock = threading.Lock()
-    
+
         def _set_nothreads(self, value, hasattr = hasattr, tuple = tuple, getattr = getattr):
             """
             Like set(), but assuming no threading is involved. It won't wake waiting threads,
@@ -861,11 +908,14 @@ except ImportError:
             if hasattr(self, '_value'):
                 # No setting twice
                 return
-            
+
             self._value = value
-    
+            self._running = False
+
             if self._cb:
-                for cb in list(self._cb):
+                cbs = list(self._cb)
+                del self._cb[:len(cbs)]
+                for cb in cbs:
                     try:
                         cb(value)
                     except:
@@ -874,8 +924,7 @@ except ImportError:
                         else:
                             error = logging.error
                         error("Error in async callback", exc_info = True)
-            self._running = False
-        
+
         def set(self, value, hasattr = hasattr, tuple = tuple, getattr = getattr):
             """
             Set the future's result as either a value, an exception wrappedn in ExceptionWrapper, or
@@ -884,13 +933,13 @@ except ImportError:
             if hasattr(self, '_value'):
                 # No setting twice
                 return
-            
+
             with self._lock:
                 old = getattr(self, '_value', None) # avoid deadlocks due to finalizers
                 cbs = list(self._cb)
                 self._value = value
             del old
-            
+
             for cb in cbs:
                 try:
                     cb(value)
@@ -901,58 +950,58 @@ except ImportError:
                         error = logging.error
                     error("Error in async callback", exc_info = True)
             self._running = False
-            
+
             event = getattr(self, '_done_event', None)
             if event is not None:
                 # wake up waiting threads
                 event.set()
-    
+
         set_result = set
-    
+
         def miss(self):
             """
             Shorthand for setting a cache miss result
             """
             self.set(CacheMissError)
-    
+
         def _miss_nothreads(self):
             """
             Shorthand for setting a cache miss result without thread safety.
             See _set_nothreads
             """
             self._set_nothreads(CacheMissError)
-    
+
         def exc(self, exc_info):
             """
             Shorthand for setting an exception result from an exc_info tuple
             as returned by sys.exc_info()
             """
             self.set(ExceptionWrapper(exc_info))
-    
+
         def _exc_nothreads(self, exc_info):
             """
             Shorthand for setting an exception result from an exc_info tuple
-            as returned by sys.exc_info(), without thread safety. 
+            as returned by sys.exc_info(), without thread safety.
             See _set_nothreads
             """
             self._set_nothreads(ExceptionWrapper(exc_info))
-    
+
         def set_exception(self, exception):
             """
             Set the Future's exception object.
             """
             self.exc((type(exception),exception,None))
-    
+
         def on_value(self, callback):
             """
-            When and if the operation completes without exception, the callback 
+            When and if the operation completes without exception, the callback
             will be invoked with its result.
             """
             def value_callback(value):
                 if value is not CacheMissError and not isinstance(value, ExceptionWrapper):
                     return callback(value)
             return self._on_stuff(value_callback)
-    
+
         def on_miss(self, callback):
             """
             If the operation results in a cache miss, the callback will be invoked
@@ -962,7 +1011,7 @@ except ImportError:
                 if value is CacheMissError:
                     return callback()
             return self._on_stuff(miss_callback)
-    
+
         def on_exc(self, callback):
             """
             If the operation results in an exception, the callback will be invoked
@@ -972,7 +1021,7 @@ except ImportError:
                 if isinstance(value, ExceptionWrapper):
                     return callback(value.value)
             return self._on_stuff(exc_callback)
-    
+
         def on_any(self, on_value = None, on_miss = None, on_exc = None):
             """
             Handy method to set callbacks for all kinds of results, and it's actually
@@ -996,7 +1045,7 @@ except ImportError:
             """
             if not self._cb:
                 self.on_any(on_value, on_miss, on_exc)
-        
+
         def on_done(self, callback):
             """
             When the operation is done, the callback will be invoked without arguments,
@@ -1005,13 +1054,13 @@ except ImportError:
             def done_callback(value):
                 return callback()
             return self._on_stuff(done_callback)
-    
+
         def chain(self, defer):
             """
             Invoke all the callbacks of the other defer
             """
             self._on_stuff(defer.set)
-    
+
         def chain_std(self, defer):
             """
             Invoke all the callbacks of the other defer, without assuming the other
@@ -1022,7 +1071,7 @@ except ImportError:
                 functools.partial(defer.set_exception, CacheMissError()),
                 lambda value : defer.set_exception(value[1] or value[0])
             )
-    
+
         def _on_stuff(self, callback, hasattr=hasattr):
             cbap = self._cb.append
             docall = hasattr(self, '_value')
@@ -1034,7 +1083,7 @@ except ImportError:
             if docall:
                 callback(self._value)
             return self
-    
+
         def add_done_callback(self, callback):
             """
             When the operatio is done, the callback will be invoked with the
@@ -1046,32 +1095,32 @@ except ImportError:
                 if self is not None:
                     return callback(self)
             return self._on_stuff(weak_callback)
-    
+
         def done(self, hasattr=hasattr, getattr=getattr):
             """
             Return True if the operation has finished, in a result or exception or cancelled, and False if not.
             """
             return hasattr(self, '_value') or getattr(self, '_cancelled', False)
-    
+
         def running(self, getattr=getattr):
             """
             Return True if the operation is running and cannot be cancelled. False if not running
             (yet or done).
             """
             return getattr(self, '_running', False)
-    
+
         def cancelled(self, getattr=getattr):
             """
             Return True if the operation has been cancelled successfully.
             """
             return getattr(self, '_cancelled', False)
-    
+
         def cancel_pending(self, getattr=getattr):
             """
             Return True if cancel was called.
             """
             return getattr(self, '_cancel_pending', False)
-    
+
         def cancel(self, getattr=getattr):
             """
             Request cancelling of the operation. If the operation cannot be cancelled,
@@ -1082,7 +1131,7 @@ except ImportError:
             else:
                 self._cancel_pending = True
                 return True
-    
+
         def set_running_or_notify_cancelled(self, getattr=getattr):
             """
             To be invoked by executors before executing the operation. If it returns True,
@@ -1093,15 +1142,15 @@ except ImportError:
             if getattr(self, '_cancel_pending', False):
                 self._cancelled = True
                 self._running = False
-    
+
                 # Notify waiters and callbacks
-                self.set_exception(CancelledError()) 
-                
+                self.set_exception(CancelledError())
+
                 return False
             else:
                 self._running = True
                 return True
-    
+
         def result(self, timeout=None, hasattr=hasattr, getattr=getattr, isinstance=isinstance, norecurse=False):
             """
             Return the operation's result, if any. If an exception was the result, re-raise it.
@@ -1140,12 +1189,12 @@ except ImportError:
                             event = self._done_event
                     else:
                         raise TimeoutError
-    
+
         def exception(self, timeout=None):
             """
             If the operation resulted in an exception, return the exception object.
             Otherwise, return None. If the operation has been cancelled, raises CancelledError,
-            and if timeout is specified and not None, and the specified time elapses without 
+            and if timeout is specified and not None, and the specified time elapses without
             a result available, raises TimeoutError.
             """
             if hasattr(self, '_value'):
@@ -1166,7 +1215,7 @@ except ImportError:
                     raise
                 except Exception,e:
                     return e
-            
+
 
 def makeFutureWrapper(base):
     """
@@ -1194,10 +1243,10 @@ class AsyncCacheProcessorThreadPool(ThreadPool):
     class AsyncCacheProcessorThread(ThreadPool.Process):
         pass
     Process = AsyncCacheProcessorThread
-    
+
     def __init__(self, workers):
         if ThreadPool is multiprocessing.pool.ThreadPool:
-            # This patches ThreadPool, which is broken when instanced 
+            # This patches ThreadPool, which is broken when instanced
             # from inside a DummyThread (happens after forking)
             current = multiprocessing.dummy.current_process()
             if not hasattr(current, '_children'):
@@ -1206,10 +1255,10 @@ class AsyncCacheProcessorThreadPool(ThreadPool):
 
 class ProcessorStats(object):
     __slots__ = (
-        'tasks_queued', 
-        'tasks_started', 
+        'tasks_queued',
+        'tasks_started',
         'tasks_cancelled',
-        'tasks_completed', 
+        'tasks_completed',
         'wait_time_sum',
         'wait_time_min',
         'wait_time_max',
@@ -1261,7 +1310,7 @@ class AsyncCacheProcessor(object):
     instead of a value, upon which an on_value(callback)
     method will retrieve the result, if any.
 
-    If there is a cache miss, on_miss callbacks will be invoked 
+    If there is a cache miss, on_miss callbacks will be invoked
     instead, and in case of an exception, on_exc.
 
     Futures also implement Python 3's concurrent.futures.Future
@@ -1271,8 +1320,8 @@ class AsyncCacheProcessor(object):
     tasks on this processor's async processing pool (in case
     you need to do it for synchronization)
     """
-    
-    def __init__(self, workers, client, 
+
+    def __init__(self, workers, client,
             coalescence_buffer_size = 500, maxqueue = None, cleanup_cycles = 500,
             threadpool = None):
         self.client = client
@@ -1295,7 +1344,7 @@ class AsyncCacheProcessor(object):
         self.tl = threading.local()
         self.cleanup_tasks = []
         self.cleanup_cycles = cleanup_cycles
-        
+
         self._tit_tat = itertools.cycle(iter((True,False))).next
 
         self.stats = ProcessorStats()
@@ -1312,7 +1361,8 @@ class AsyncCacheProcessor(object):
         return self._threadpool
 
     def _enqueue(self, action, coalesce = None, coalesce_key = NONE):
-        cfuture = future = Future(logger=self.logger)
+        logger = self.logger
+        cfuture = future = Future(logger)
 
         do_coalescence = coalesce is not None and coalesce_key is not NONE
         if do_coalescence:
@@ -1326,10 +1376,11 @@ class AsyncCacheProcessor(object):
                 except:
                     pass
                 cfuture, cwaction = coalesce.setdefault(coalesce_key, (future, waction))
-                self.logger.warning("Found zombie task in coalesce buffer with key %r", coalesce_key)
+                logger.warning("Found zombie task in coalesce buffer with key %r", coalesce_key)
 
         if cfuture is future:
-            if self.maxqueue is not None and self.queuelen > (self.maxqueue*2):
+            maxqueue = self.maxqueue
+            if maxqueue is not None and self.queuelen > (maxqueue*2):
                 # Stop filling it
                 if do_coalescence:
                     try:
@@ -1342,36 +1393,37 @@ class AsyncCacheProcessor(object):
                 wself = self._wself
                 stats = self.stats
                 queue_time = time.time()
-                stats.on_task_queued()
                 def wrapped_action():
-                    def clean():
+                    def clean_coalesce():
                         if do_coalescence:
                             try:
                                 del coalesce[coalesce_key]
                             except:
                                 pass
-    
-                        self = wself()
+
+                    def clean(self):
                         if self is not None:
                             try:
-                                if not hasattr(self.tl, 'dirty_rounds'):
-                                    self.tl.dirty_rounds = 0
-                                self.tl.dirty_rounds += 1
-                                if self.tl.dirty_rounds > self.cleanup_cycles or not self.queuelen:
-                                    self.tl.dirty_rounds = 0
+                                tl = self.tl
+                                if not hasattr(tl, 'dirty_rounds'):
+                                    tl.dirty_rounds = 0
+                                tl.dirty_rounds += 1
+                                if tl.dirty_rounds > self.cleanup_cycles or not self.queuelen:
+                                    tl.dirty_rounds = 0
                                     for task in self.cleanup_tasks:
                                         task()
                                     for task in _global_cleanup_tasks:
                                         task()
                             except:
-                                self.logger.error("Error during background thread cleanup", exc_info = True)
-                    
+                                logger.error("Error during background thread cleanup", exc_info = True)
+
                     # discard queue head quickly when we're overloaded
                     # head is always less relevant
                     self = wself()
+                    cleaned_coalesce = False
                     try:
-                        if self is not None and self.maxqueue is not None:
-                            if self.queuelen > self.maxqueue:
+                        if self is not None and maxqueue is not None:
+                            if self.queuelen > maxqueue:
                                 # Only discard half the entries, otherwise we can
                                 # enter a race condition in which really fast input
                                 # to the processor keeps the queue full and discarding all
@@ -1383,24 +1435,31 @@ class AsyncCacheProcessor(object):
                             stats.on_task_started(start_time - queue_time)
                             try:
                                 rv = action()
-                                clean()
+                                clean_coalesce()
+                                cleaned_coalesce = True
                                 cfuture.set(rv)
                             except CacheMissError:
-                                clean()
+                                clean_coalesce()
+                                cleaned_coalesce = True
                                 cfuture.miss()
                             except:
-                                clean()
+                                clean_coalesce()
+                                cleaned_coalesce = True
                                 # Clear up traceback to avoid leaks
                                 cfuture.exc(sys.exc_info()[:-1] + (None,))
                             stats.on_task_completed(time.time() - start_time)
                         else:
+                            clean_coalesce()
+                            cleaned_coalesce = True
                             stats.on_task_cancelled()
-                            clean()
-                    except:
+                    finally:
                         # Just in case, we really need to clean, or we leak cfutures
-                        clean()
-                        raise
+                        if not cleaned_coalesce:
+                            clean_coalesce()
+                            cleaned_coalesce = True
+                        clean(self)
                 self.threadpool.apply_async(wrapped_action, ())
+                stats.on_task_queued()
         return cfuture
 
     @property
@@ -1421,7 +1480,7 @@ class AsyncCacheProcessor(object):
     def do_async_coalescent(self, coalesce, coalesce_key, func, *args, **kwargs):
         """
         Returns an already-queued future if there is one, or a new one if not.
-        
+
         Params:
             coalesce: A coalescence map (dict, Cache, or simialr)
             coalesce_key: A coalesence key
@@ -1436,7 +1495,7 @@ class AsyncCacheProcessor(object):
         Returns a proxy of this processor bound to the specified client instead.
         """
         return WrappedCacheProcessor(self, client)
-    
+
     def getTtl(self, key, default = NONE, **kw):
         if not kw or not (kw.viewkeys() - COALESCE_IGNORE_KWARGS):
             if default is NONE:
@@ -1447,7 +1506,7 @@ class AsyncCacheProcessor(object):
             ckey = NONE
         return self._enqueue(functools.partial(self.client.getTtl, key, default, **kw),
             self.coalesce_getTtl, ckey)
-    
+
     def get(self, key, default = NONE, **kw):
         if not kw or not (kw.viewkeys() - COALESCE_IGNORE_KWARGS):
             if default is NONE:
@@ -1458,7 +1517,7 @@ class AsyncCacheProcessor(object):
             ckey = NONE
         return self._enqueue(functools.partial(self.client.get, key, default, **kw),
             self.coalesce_get, ckey)
-    
+
     def contains(self, key, *p, **kw):
         if not p and not kw:
             ckey = key
@@ -1489,7 +1548,7 @@ class AsyncCacheProcessor(object):
 
     def purge(self):
         return self._enqueue(self.client.purge)
-    
+
     def register_cleanup(self, task):
         """
         Register a callable that will take no arguments, and will
@@ -1548,7 +1607,7 @@ class WrappedCacheProcessor(object):
             return self
         else:
             return WrappedCacheProcessor(self.processor, client)
-    
+
     def getTtl(self, key, default = NONE, **kw):
         if not kw:
             if default is NONE:
@@ -1557,9 +1616,9 @@ class WrappedCacheProcessor(object):
                 ckey = (key, default)
         else:
             ckey = NONE
-        return self.processor.do_async_coalescent(self.coalesce_getTtl, ckey, 
+        return self.processor.do_async_coalescent(self.coalesce_getTtl, ckey,
             self.client.getTtl, key, default, **kw)
-    
+
     def get(self, key, default = NONE):
         if default is NONE:
             ckey = key
@@ -1567,7 +1626,7 @@ class WrappedCacheProcessor(object):
             ckey = (key, default)
         return self.processor.do_async_coalescent(self.coalesce_get, ckey,
             self.client.get, key, default)
-    
+
     def contains(self, key, *p, **kw):
         if not p and not kw:
             ckey = key
@@ -1598,7 +1657,7 @@ class WrappedCacheProcessor(object):
 
     def purge(self, *p, **kw):
         return self.processor.do_async(self.client.purge, *p, **kw)
-    
+
     def register_cleanup(self, task):
         self.processor.register_cleanup(task)
 
