@@ -26,12 +26,17 @@ class WorkerThread(threading.Thread):
         self.kwargs = kwargs
         self.__terminate = False
 
-    def run(self):
+    def run(self, TerminateWorker=TerminateWorker):
         while not self.__terminate:
             try:
                 try:
                     self.target(*self.args,**self.kwargs)
                 finally:
+                    worker_ = worker
+                    if worker_ is None:
+                        # The None check is necessary to avoid errors during interpreter shutdown
+                        self.logger.info("Interpreter shutdown in progress, terminating worker thread")
+                        raise TerminateWorker()
                     worker._callCleanupHooks()
             except TerminateWorker:
                 self.logger.info("Worker terminated")
@@ -63,16 +68,18 @@ except ImportError:
             raise exc[0], exc[1], exc[2]
 
 class WaitIter:
-    def __init__(self, event, queues, timeout = 5):
+    def __init__(self, event, timeout = None):
         self.event = event
-        self.queues = queues
         self.timeout = timeout
-        self.terminate = False
+        self._terminate = False
+    def terminate(self):
+        self._terminate = True
+        self.event.set()
     def __iter__(self):
         return self
-    def next(self):
-        if self.terminate:
-            threading.current_thread().terminate(False)
+    def next(self, TerminateWorker=TerminateWorker):
+        if self._terminate:
+            raise TerminateWorker()
         self.event.wait(self.timeout)
         raise StopIteration
 
@@ -84,17 +91,17 @@ class ThreadPool:
     multiple queues for managing fairness.
 
     This implementation is forcibly a daemon thread pool, which when destroyed
-    will cancel all pending tasks, and no task returns any result, and has been 
+    will cancel all pending tasks, and no task returns any result, and has been
     optimized for that usage pattern.
     """
-    
+
     Process = WorkerThread
 
     def __init__(self, workers = None, min_batch = 10, max_batch = 1000, max_slice = None, logger = None,
             name_pattern = None):
         if workers is None:
             workers = multiprocessing.cpu_count()
-        
+
         self.workers = workers
         self.name_pattern = name_pattern
         self.logger = logger if logger is not None else logging.getLogger('chorde')
@@ -107,7 +114,7 @@ class ThreadPool:
         self.__empty = threading.Event()
         self.__empty.set()
         self.__cleanup_callbacks = []
-        
+
         self.local = threading.local()
         self.queues = collections.defaultdict(list)
         self.queue_weights = {}
@@ -117,7 +124,7 @@ class ThreadPool:
         self.__workset = set()
         self.__busyqueues = set()
         self.__busyfactors = {}
-        self.__exhausted_iter = WaitIter(self.__not_empty, self.queues)
+        self.__exhausted_iter = WaitIter(self.__not_empty)
         self.__dequeue = self.__exhausted = self.__exhausted_iter.next
 
         self.min_batch = min_batch
@@ -127,7 +134,7 @@ class ThreadPool:
     def queuelen(self, queue = None):
         return (
             len(self.queues.get(queue,()))
-            - self.__queue_slices.get(queue,0) 
+            - self.__queue_slices.get(queue,0)
             + int(self.__worklen * self.__busyfactors.get(queue,0))
         )
 
@@ -156,19 +163,17 @@ class ThreadPool:
                 self.logger.error("Error in task cleanup callback", exc_info = True)
 
     def __swap_queues(self, max=max, min=min, len=len):
-        qpop = self.queues.pop
-        qget = self.queues.get
+        queues = self.queues
+        qget = queues.get
         queue_slices = self.__queue_slices
         pget = queue_slices.get
         ppop = queue_slices.pop
         qprio = self.queue_weights.get
-        qnames = self.queues.keys()
+        qnames = queues.keys()
         wqueues = []
         wprios = []
-        wposes = []
         iquantities = {}
         itotal = 0
-        can_straggle = False
 
         if qnames:
             # Compute batch size
@@ -181,53 +186,41 @@ class ThreadPool:
             min_batch = self.min_batch
             max_batch = self.max_batch
             max_slice = self.max_slice
-            qslots = min(max_batch, max(min_batch,min([len(qget(q)) / qprio(q,1) for q in qnames])))
+            qslots = min(max_batch, max(min_batch, min([
+                (len(qget(q)) or max_batch) / qprio(q,1)
+                for q in qnames
+            ])))
             for qname in qnames:
                 q = qget(qname)
+                if not q:
+                    continue
+
                 qpos = pget(qname,0)
                 prio = qprio(qname,1)
-                margin = max(prio,min_batch)
                 batch = qslots * prio
-                if batch >= (len(q) - margin - qpos):
-                    #print "move %s" % (qname,)
-                    q = qpop(qname)
-                    if qpos:
-                        del q[:qpos] # atomic re. pushes
-                        ppop(qname,None) # reset
-                    wqueues.append(q)
-                    wposes.append(0)
-                    qlen = len(q)
+
+                # copy-slicing
+                qslice = q[qpos:qpos+batch]
+                qlen = len(qslice)
+                if qlen:
                     iquantities[qname] = qlen
                     itotal += qlen
-                    can_straggle = True
-                else:
-                    if qpos > (max_slice or (len(q)/2)):
-                        # copy-slicing
-                        #print "copy-slice %s[%d:%d] of %d" % (qname,qpos,qpos+batch,len(q))
-                        qslice = q[qpos:qpos+batch]
-                        qlen = len(qslice)
-                        iquantities[qname] = qlen
-                        itotal += qlen
-                        wqueues.append(qslice)
-                        del q[:qpos+batch]
-                        del qslice
+                    wqueues.append(qslice)
+                    wprios.append(prio)
+                del qslice
+
+                if qlen < batch or qpos > (max_slice or (len(q)/2)):
+                    del q[:qpos+qlen]
+                    if qpos:
                         ppop(qname,None)
-                        wposes.append(0)
-                    else:
-                        # zero-copy slicing
-                        #print "iter-slice %s[%d:%d] of %d" % (qname,qpos,qpos+batch,len(q))
-                        qlen = min(batch, max(1, len(q) - qpos))
-                        iquantities[qname] = qlen
-                        itotal += qlen
-                        wqueues.append(itertools.islice(q, qpos, qpos+batch)) # queue heads are immutable
-                        queue_slices[qname] = qpos+batch
-                        wposes.append(None)
-                wprios.append(prio)
+                else:
+                    # move slice position
+                    queue_slices[qname] = qpos+qlen
 
         if wqueues:
             self.__busyqueues.clear()
             self.__busyqueues.update(qnames)
-            
+
             # Flatten with weights
             # Do it repeatedly to catch stragglers (those that straggle past the flattening step)
             iqueue = []
@@ -237,38 +230,20 @@ class ThreadPool:
             izip = itertools.izip
             repeat = itertools.repeat
             partial = functools.partial
-            retry = True
 
-            while retry:
-                if can_straggle:
-                    # Wait for stragglers
-                    time.sleep(0.0001)
-                
-                queues = []
-                qposes = []
-                for q,qprio,wpos in izip(wqueues, wprios, wposes):
-                    if wpos is not None:
-                        # must slice to make sure we take a stable snapshot of the list
-                        # we'll process stragglers on the next iteration
-                        qlen = len(q)
-                        qiter = iter(islice(q,wpos,wpos+qlen))
-                        qposes.append(wpos+qlen)
-                    else:
-                        qiter = iter(q)
-                        qposes.append(None)
-                    queues.append(partial(repeat, qiter.next, qprio))
-                wposes = qposes
-                
-                ioffs = 0
-                ilen = len(iqueue)
-                while queues:
-                    try:
-                        for ioffs,q in islice(cycle(enumerate(queues)), ioffs, None):
-                            for q in q():
-                                iappend(q())
-                    except StopIteration:
-                        del queues[ioffs]
-                retry = can_straggle and len(iqueue) != ilen
+            queues = []
+            for q,qprio in izip(wqueues, wprios):
+                queues.append(partial(repeat, iter(q).next, qprio))
+
+            ioffs = 0
+            while queues:
+                try:
+                    for ioffs,q in islice(cycle(enumerate(queues)), ioffs, None):
+                        for q in q():
+                            iappend(q())
+                except StopIteration:
+                    del queues[ioffs]
+
             self.__worklen = len(iqueue)
             self.__dequeue = iter(iqueue).next
             if itotal:
@@ -290,27 +265,28 @@ class ThreadPool:
         elif self.__not_empty.isSet():
             # Try again
             # Someone may have sneaked in while we were in the above case
+            # and may still neak in after we clear the not_empty event
             self.__not_empty.clear()
             self.__worklen = 0
             self.__busyfactors = {}
             self.__dequeue = self.__exhausted
             self.__swap_queues()
         else:
-            # Still empty, give up
-            self.__not_empty.clear()
-            self.__worklen = 0
-            self.__busyfactors = {}
-            self.__dequeue = self.__exhausted
+            # Still empty, can safely give up until signaled
+            # Wake up threads trying to join
+            self.__empty.set()
 
-    def _dequeue(self):
+    def _dequeue(self, TerminateWorker=TerminateWorker):
         tid = thread.get_ident()
         workset = self.__workset
+        termcount = 0
         while True:
-            if self.__dequeue is self.__exhausted and not self.queues and not self.__not_empty.is_set():
+            if (self.__dequeue is self.__exhausted and not self.__not_empty.isSet()
+                    and (not self.queues or not any(self.queues.values()))):
                 # Sounds like there's nothing to do
                 # Yeah, gonna wait
                 workset.discard(tid)
-                if not workset and not self.queues:
+                if not workset and (not self.queues or not any(self.queues.values())):
                     self.__empty.set()
             else:
                 workset.add(tid)
@@ -318,13 +294,21 @@ class ThreadPool:
                 rv = self.__dequeue()
                 self.__worklen -= 1 # not atomic, but we don't care
                 return rv
-            except StopIteration:
+            except (TerminateWorker, StopIteration) as e:
                 # Exhausted whole workqueue?
                 with self.__swap_lock:
                     try:
                         if self.__dequeue is self.__exhausted:
                             # Pointless to wait, just swap again
-                            raise StopIteration
+                            if termcount > 0:
+                                raise TerminateWorker()
+                            else:
+                                # First time we get the terminate signal, we try to swap queues
+                                # to flush any queued tasks. If we get 2 consecutive signals,
+                                # that means we're done with the queue.
+                                if isinstance(e, TerminateWorker):
+                                    termcount += 1
+                                raise StopIteration
                         else:
                             # Try it
                             workset.add(tid)
@@ -334,32 +318,36 @@ class ThreadPool:
                     except StopIteration:
                         # Yep, exhausted queue, build up new workqueue
                         self.__swap_queues()
+                    except TerminateWorker:
+                        # Wake up others so they check, if they're sleeping
+                        workset.discard(tid)
+                        self.__not_empty.set()
+                        raise
 
     def _enqueue(self, queue, task):
         self.queues[queue].append(task)
-        if self.__dequeue is self.__exhausted:
-            # Wake up waiting threads
-            # Note that it's not necessary if dequeue is set to a dequeuing
-            # iterator, since that means threads are busy working already
-            # It is also not necessary to invoke this all the time. If the
-            # flags are the right way at any point within this function being
-            # run, then it already means the respective waiting threads have
-            # woken up (or are in the process of waking up) in time to pick up the 
-            # just-queued value, so avoid the actual operation 
-            # (which is much more expensive than checking)
-            not_empty = self.__not_empty
-            if not not_empty.isSet():
-                not_empty.set()
-            empty = self.__empty
-            if empty.isSet():
-                empty.clear()
+
+        # Wake up waiting threads
+        # Note that it's not necessary to invoke this all the time. If the
+        # flags are the right way at any point within this function being
+        # run, then it already means the respective waiting threads have
+        # woken up (or are in the process of waking up) in time to pick up the
+        # just-queued value, so avoid the actual operation
+        # (which is much more expensive than checking)
+        not_empty = self.__not_empty
+        if not not_empty.isSet():
+            not_empty.set()
+        empty = self.__empty
+        if empty.isSet():
+            empty.clear()
+
         self.assert_started()
 
     @staticmethod
-    def worker(self):
+    def worker(self, TerminateWorker=TerminateWorker):
         self = self()
         if self is None:
-            raise TerminateWorker
+            raise TerminateWorker()
 
         task = self._dequeue()
         local = self.local
@@ -402,9 +390,12 @@ class ThreadPool:
                         pass
                 self.__workers = None
 
+            # Wake up threads so they die awake
+            self.__not_empty.set()
+
     def close(self):
         # Signal idle threads to commit suicide
-        self.__exhausted_iter.terminate = True
+        self.__exhausted_iter.terminate()
 
     def terminate(self):
         self.stop()
@@ -418,24 +409,41 @@ class ThreadPool:
             self.populate_workers()
 
     def join(self, timeout = None):
+        if not self.is_started():
+            return True
         if timeout is not None:
             now = time.time()
             timeout += now
         while timeout is None or now < timeout:
-            if self.__empty.wait(timeout - now if timeout is not None else None):
+            if timeout is not None:
+                wait_timeout = timeout - now
+            elif self.__exhausted_iter._terminate:
+                # If the pool is shut down in a way that aborts all queued tasks, we can't
+                # merely wait for the empty event to be set, we have to also monitor suiciding threads
+                if not self.is_started():
+                    break
+                wait_timeout = 0.1
+            else:
+                wait_timeout = None
+            if self.__empty.wait(wait_timeout):
                 # The event is not 100% certain, we can still get awakened when there's work to do
                 # We have to check under __swap_lock to be sure
                 with self.__swap_lock:
-                    if self.__dequeue is self.__exhausted and not self.queues and not self.__workset:
+                    if (self.__dequeue is self.__exhausted and not self.__workset
+                            and (not self.queues or not any(self.queues.values()))):
                         return True
                     else:
                         # False alarm, clear it so we don't spin
                         self.__empty.clear()
             else:
                 # Timeout
-                return False
+                if timeout is not None or not self.__workset:
+                    return (self.__dequeue is self.__exhausted and not self.__workset
+                        and (not self.queues or not any(self.queues.values())))
             if timeout is not None:
                 now = time.time()
+        return not self.is_started() or (self.__dequeue is self.__exhausted and not self.__workset
+            and (not self.queues or not any(self.queues.values())))
 
     def populate_workers(self):
         with self.__spawnlock:
@@ -451,10 +459,11 @@ class ThreadPool:
                     w.logger = self.logger
                     w.daemon = True
                     w.start()
-                
+
                 self.__pid = os.getpid()
             # Else, just keep number of workers in sync
             elif len(self.__workers) < self.workers:
+                name_pattern = self.name_pattern
                 new_workers = self.workers - len(self.__workers)
                 index_base = self.__last_worker_index
                 self.__last_worker_index += new_workers
@@ -507,10 +516,10 @@ class SubqueueWrapperThreadPool:
     multiple queues for managing fairness.
 
     This implementation is forcibly a daemon thread pool, which when destroyed
-    will cancel all pending tasks, and no task returns any result, and has been 
+    will cancel all pending tasks, and no task returns any result, and has been
     optimized for that usage pattern.
     """
-    
+
     def __init__(self, pool, queue, priority = None):
         self.queue = queue
         self.pool = pool
